import configparser
import itertools
import json
import os
import time
from pathlib import Path

from colorama import Fore, Style
import dill
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from grid2op.Episode import EpisodeData
from grid2op.PlotGrid import PlotPlotly, PlotMatplot

from grid2viz.src.kpi.EpisodeAnalytics import EpisodeAnalytics

# refer to https://github.com/rte-france/Grid2Op/blob/master/getting_started/8_PlottingCapabilities.ipynb for better usage

graph = None
graph_matplotlib = None


# TODO: addSubstationColor - integrate that into grid2op Plotgrid
def add_substation_color_matplot(subs, plot_helper, fig):
    radius_size = plot_helper._sub_radius
    # fig = plot_helper.plot_layout()
    ax = fig.gca()

    for id_sub in subs:
        subName = "sub_" + str(id_sub)
        x, y = plot_helper._grid_layout[subName]
        circle = plt.Circle((x, y), int(radius_size), color="gold")
        ax.add_artist(circle)

    return fig


def add_substation_color_plotly(subs, plot_helper, fig, color="gold"):
    radius_size = int(plot_helper._sub_radius * 0.8)

    for id_sub in subs:
        subName = "sub_" + str(id_sub)
        x_center, y_center = plot_helper._grid_layout[subName]

        marker_dict = dict(
            size=radius_size,
            color=color,
            showscale=False,
            opacity=0.5,
        )
        fig.add_trace(
            go.Scatter(
                x=[x_center],
                y=[y_center],
                mode="markers",
                text=[subName],
                name="sub" + subName,
                marker=marker_dict,
                showlegend=False,
            )
        )
    return fig


def make_network(episode, responsive=True):
    """
    Create a Plotly network graph with the layout configuration and the selected episode.

    :param episode: An episode containing targeted data for the graph.
    :return: Network graph
    """
    global graph
    if graph is None:
        graph = PlotPlotly(
            grid_layout=episode.observation_space.grid_layout,
            observation_space=episode.observation_space,
            responsive=responsive,
        )
    return graph


def make_network_matplotlib(episode):
    global graph_matplotlib
    if graph_matplotlib is None:
        graph_matplotlib = PlotMatplot(
            grid_layout=episode.observation_space.grid_layout,
            observation_space=episode.observation_space,
            line_name=False,
            gen_name=False,
            load_name=False,
        )
    return graph_matplotlib


######
# we want a non responsive graph for now in agent_study
# so we have to define it differently from the global graph in make_network that we don't use here
def make_network_agent_study(episode, timestep):
    # subs_on_bus_2 = np.repeat(False, episode_data.observations[0].n_sub)
    graph = PlotPlotly(
        grid_layout=episode.observation_space.grid_layout,
        observation_space=episode.observation_space,
        responsive=False,
    )
    graph._sub_radius = 30  # instead of 25 by default
    graph._bus_radius = 10  # instead of 4 by default

    fig = graph.plot_obs(episode.observations[timestep])

    ##########
    # We color subs where we had actions
    sub_name_modified = list(
        itertools.chain.from_iterable(episode.action_data_table.subs_modified)
    )
    sub_id_modified = [
        int(str.split("_")[1])
        for str in episode.action_data_table.subs_modified[timestep]
    ]
    fig = add_substation_color_plotly(sub_id_modified, graph, fig)

    # coloring subs not in reference topologie
    nb_bus_subs = [
        episode.observations[timestep].state_of(substation_id=i)["nb_bus"]
        for i in range(episode.observations[timestep].n_sub)
    ]
    sub_2buses = [
        i for i in range(episode.observations[timestep].n_sub) if nb_bus_subs[i] >= 2
    ]
    fig = add_substation_color_plotly(
        sub_2buses, graph, fig, color="green"
    )  # also other color for subs not in ref topo
    return fig


def make_network_agent_overview(episode):
    graph = make_network(episode)

    # modified_lines = actions_model.get_modified_lines(episode)
    # line_values = [None] * episode.n_lines
    # for line in modified_lines.index:
    #    line_values[np.where(episode.line_names == line)[0][0]] = line

    lines_attacked = list(
        episode.attacks_data_table["id_lines"][
            episode.attacks_data_table.attack
        ].unique()
    )
    lines_overflowed_ids = list(
        itertools.chain.from_iterable(episode.total_overflow_ts.line_ids)
    )
    # to color assets on our graph with different colors while not overloading it with information
    # we will use plot_obs instead of plot_info for now
    ####
    # For that we override an observation with the desired values
    obs_colored = episode.observations[0]

    # having a rho with value 1.0 give us a red line while 0.7 gives us an orange line and 0.3 a blue line
    rho_to_color = np.array(
        [
            float(0.6) if line in lines_attacked else float(0.3)
            for line in episode.line_names
        ]
    )
    rho_to_color[lines_overflowed_ids] = 1.0
    line_status_colored = np.array(
        [False if line in lines_attacked else True for line in episode.line_names]
    )
    obs_colored.rho = rho_to_color
    obs_colored.line_status = line_status_colored

    # network_graph = make_network(episode).plot_info(
    #    line_values=[ line if line in lines_attacked else None for line in  episode.line_names]
    #    #coloring="line"
    # )
    # )
    fig = graph.plot_obs(obs_colored, line_info=None, gen_info=None, load_info=None)

    ##########
    # We color subs where we had actions
    sub_name_modified = list(
        itertools.chain.from_iterable(episode.action_data_table.subs_modified)
    )
    sub_id_modified = set([int(str.split("_")[1]) for str in sub_name_modified])
    fig = add_substation_color_plotly(sub_id_modified, graph, fig)

    return fig


def make_network_scenario_overview(episode):
    max_loads = (
        episode.load[["value", "equipement_id"]]
        .groupby("equipement_id")
        .max()
        .sort_index()
    )
    max_gens = (
        episode.production[["value", "equipement_id"]]
        .groupby("equipement_id")
        .max()
        .sort_index()
    )
    lines_in_maintenance = list(
        episode.maintenances["line_name"][episode.maintenances.value == 1].unique()
    )

    graph = make_network_matplotlib(episode)

    # to color assets on our graph with different colors while not overloading it with information
    # we will use plot_obs instead of plot_info for now
    ####
    # For that we override an observation with the desired values
    obs_colored = episode.observations[0]

    # having a rho with value 0.1 give us a blue line while 0.5 gives us an orange line
    # line in maintenance would display as dashed lines
    rho_to_color = np.array(
        [
            float(0.0) if line in lines_in_maintenance else float(0.4)
            for line in episode.line_names
        ]
    )
    line_status_colored = np.array(
        [False if line in lines_in_maintenance else True for line in episode.line_names]
    )
    obs_colored.rho = rho_to_color
    obs_colored.line_status = line_status_colored

    obs_colored.load_p = np.array(max_loads.value)
    obs_colored.prod_p = np.array(max_gens.value)

    network_graph = graph.plot_obs(obs_colored, line_info=None)
    # network_graph=graph.plot_info(
    #    #observation=episode.observations[0],
    #    load_values=max_loads.values.flatten(),
    #    load_unit="MW",
    #    gen_values=max_gens.values.flatten(),
    #    gen_unit="MW"
    #    #line_values=[ 1 if line in lines_in_maintenance else 0 for line in  episode.line_names],
    #    #coloring="line"
    # )

    return network_graph


store = {}


def make_episode(agent, episode_name):
    """
    Load episode from cache. If not already in, compute episode data
    and save it in cache.

    :param agent: Agent Name
    :param episode_name: Name of the studied episode
    :return: Episode with computed data
    """
    if is_in_ram_cache(episode_name, agent):
        return get_from_ram_cache(episode_name, agent)
    elif is_in_fs_cache(episode_name, agent):
        episode = get_from_fs_cache(episode_name, agent)
        save_in_ram_cache(episode_name, agent, episode)
        return episode
    else:
        episode = compute_episode(episode_name, agent)
        save_in_ram_cache(episode_name, agent, episode)
        return episode


def make_episode_without_decorate(agent, episode_name):
    """
    Load episode from cache without decorating with the EpisodeData attributes
    This is needed to use multiprocessing which pickles/unpickles the results.

    :param agent: Agent Name
    :param episode_name: Name of the studied episode
    :return: Episode with computed data (without EpisodeData attributes), EpisodeData instance
    """
    if is_in_ram_cache(episode_name, agent):
        return get_from_ram_cache(episode_name, agent)
    elif is_in_fs_cache(episode_name, agent):
        beg = time.time()
        path = get_fs_cached_file(episode_name, agent)
        print(
            f"Loading from filesystem cache agent {agent} on scenario {episode_name}..."
        )
        with open(path, "rb") as f:
            episode_analytics = dill.load(f)
        end = time.time()
        print(
            f"Agent {agent} on scenario {episode_name} loaded from filesystem cache in: {(end - beg):.1f} s"
        )
        return episode_analytics
    else:
        episode_data = retrieve_episode_from_disk(episode_name, agent)
        if episode_data is not None:
            episode_analytics = EpisodeAnalytics(episode_data, episode_name, agent)
            save_in_fs_cache(episode_name, agent, episode_analytics)
            return episode_analytics
        else:
            return None


def clear_fs_cache():
    os.rmdir(cache_dir)


def is_in_fs_cache(episode_name, agent):
    return os.path.isfile(get_fs_cached_file(episode_name, agent))


def get_fs_cached_file(episode_name, agent):
    episode_dir = os.path.join(cache_dir, episode_name)
    if not os.path.exists(episode_dir):
        os.makedirs(episode_dir)
    return os.path.join(episode_dir, agent + ".dill")


def save_in_fs_cache(episode_name, agent, episode):
    path = get_fs_cached_file(episode_name, agent)
    with open(path, "wb") as f:
        dill.dump(episode, f, protocol=4)


def get_from_fs_cache(episode_name, agent):
    beg = time.time()
    path = get_fs_cached_file(episode_name, agent)
    print(f"Loading from filesystem cache agent {agent} on scenario {episode_name}...")
    episode_data = retrieve_episode_from_disk(episode_name, agent)
    with open(path, "rb") as f:
        episode_analytics = dill.load(f)

    episode_analytics.decorate(episode_data)
    end = time.time()
    print(
        f"Agent {agent} on scenario {episode_name} loaded from filesystem cache in: {(end - beg):.1f} s"
    )
    return episode_analytics


def compute_episode(episode_name, agent):
    print(f"Loading from logs agent {agent} on scenario {episode_name}...")
    beg = time.time()
    episode_data = retrieve_episode_from_disk(episode_name, agent)
    episode_analytics = EpisodeAnalytics(episode_data, episode_name, agent)
    save_in_fs_cache(episode_name, agent, episode_analytics)
    episode_analytics.decorate(episode_data)
    end = time.time()
    print(
        f"Agent {agent} on scenario {episode_name} loaded from logs in: {(end - beg):.1f} s"
    )
    return episode_analytics


def retrieve_episode_from_disk(episode_name, agent):
    path = os.path.join(agents_dir, agent)
    episode_path = os.path.abspath(os.path.join(path, episode_name))
    if os.path.isdir(episode_path):
        episode_data = EpisodeData.from_disk(path, episode_name)
        return episode_data
    else:
        return None


def is_in_ram_cache(episode_name, agent):
    return make_ram_cache_id(episode_name, agent) in store


def save_in_ram_cache(episode_name, agent, episode):
    store[make_ram_cache_id(episode_name, agent)] = episode


def get_from_ram_cache(episode_name, agent):
    return store[make_ram_cache_id(episode_name, agent)]


def make_ram_cache_id(episode_name, agent):
    return agent + episode_name


def check_all_tree_and_get_meta_and_best(base_dir, agents):
    best_agents = {}
    meta_json = {}
    scenarios = set()
    survival_dic = {}

    for agent in agents:
        survival_dic_agent = {}
        for scenario_name in os.listdir(os.path.join(base_dir, agent)):

            scenario_folder = os.path.join(base_dir, agent, scenario_name)
            if not os.path.isdir(scenario_folder):
                continue
            with open(os.path.join(scenario_folder, "episode_meta.json")) as f:
                episode_meta = json.load(fp=f)
                meta_json[scenario_name] = episode_meta

                survival_dic_agent[scenario_name] = int(
                    int(episode_meta["nb_timestep_played"])
                    * 100
                    / int(episode_meta["chronics_max_timestep"])
                )
                scenarios.add(scenario_name)

                if scenario_name not in best_agents:
                    best_agents[scenario_name] = {
                        "value": -1,
                        "agent": None,
                        "out_of": 0,
                        "cum_reward": -1,
                    }
                condition_to_update_best_agent = best_agents[scenario_name][
                    "value"
                ] < episode_meta["nb_timestep_played"] or (
                    best_agents[scenario_name]["value"]
                    == episode_meta["nb_timestep_played"]
                    and best_agents[scenario_name]["cum_reward"]
                    < episode_meta["cumulative_reward"]
                )
                if condition_to_update_best_agent:
                    best_agents[scenario_name]["value"] = episode_meta[
                        "nb_timestep_played"
                    ]
                    best_agents[scenario_name]["agent"] = agent
                    best_agents[scenario_name]["cum_reward"] = episode_meta[
                        "cumulative_reward"
                    ]

            best_agents[scenario_name]["out_of"] = (
                best_agents[scenario_name]["out_of"] + 1
            )
        survival_dic[agent] = survival_dic_agent

    survival_df = pd.DataFrame(columns=agents, index=scenarios)
    for agent in agents:
        survival_dic_agent = survival_dic[agent]
        for (scenario, survival_time) in survival_dic_agent.items():
            survival_df.loc[scenario][agent] = survival_time

    survival_df = survival_df.fillna(-1)  # To be able to cast as int below.
    survival_df = survival_df.astype(int)

    return meta_json, best_agents, survival_df


"""
Initialisation routine
"""
""" Parsing of config file"""
path_cfg = os.path.join(os.environ["GRID2VIZ_ROOT"], "config.ini")
parser = configparser.ConfigParser()
print(
    Fore.BLUE + Style.BRIGHT + "The config file used is located at: {}".format(path_cfg)
)
parser.read(path_cfg)

agents_dir = parser.get("DEFAULT", "agents_dir")
print(Fore.BLUE + "Agents data used is located at: {}".format(agents_dir))
cache_dir = os.path.join(agents_dir, "_cache")
"""Parsing of agent folder tree"""
agents = sorted(
    [
        file
        for file in os.listdir(agents_dir)
        if os.path.isdir(os.path.join(agents_dir, file)) and not file.startswith("_")
    ]
)
meta_json, best_agents, survival_df = check_all_tree_and_get_meta_and_best(
    agents_dir, agents
)
scenarios = []
scenarios_agent = {}
agent_scenario = {}

n_cores = int(parser.get("DEFAULT", "n_cores"))

for agent in agents:
    scen_path = os.path.join(agents_dir, agent)
    scens = [
        file
        for file in os.listdir(scen_path)
        if os.path.isdir(os.path.join(scen_path, file))
    ]
    scenarios_agent[agent] = scens
    for scen in scens:
        if scen not in agent_scenario:
            agent_scenario[scen] = []
        if agent not in agent_scenario[scen]:
            agent_scenario[scen].append(agent)
    scenarios = scenarios + scens

scenarios = set(scenarios)

# Create a .grid2viz directory in the user home directory
grid2viz_home_directory = Path.home() / ".grid2viz"
grid2viz_home_directory.mkdir(parents=False, exist_ok=True)



"""
    Utility functions for creation of graph and graph component used several times.
"""

import numpy as np
import pandas as pd
from plotly import graph_objects as go

from grid2viz.src.kpi import EpisodeTrace, observation_model
from grid2viz.src.kpi.actions_model import get_actions_sum
from grid2viz.src.manager import make_episode


def ts_graph_avail_assets(ts_kind, episode):
    """
    Get a list of available assets for a selected kind of timeserie of an episode.

    This method is used with the followed kind of timeseries:
        - Hazards
        - Maintenances
        - Production
        - Load

    .. warning:: By default if no kind is given the load's assets will be returned

    :param ts_kind: kind of timeseries wanted
    :param episode: episode studied
    :param prod_types: The different types of production
    :return: list of of available asset in form of tuple (option, value)
    """
    if ts_kind in ["Hazards", "Maintenances"]:
        options, value = [
            {"label": line_name, "value": line_name}
            for line_name in ["total", *episode.line_names]
        ], "total"  # episode.line_names[0]
    elif ts_kind == "Production":
        prod_types = episode.get_prod_types()
        options = [
            {"label": prod_name, "value": prod_name}
            for prod_name in [
                *list(set(prod_types.values())),
                "total",
                *episode.prod_names,
            ]
        ]
        value = "solar"  # episode.prod_names[0]
    else:
        options = [
            {"label": load_name, "value": load_name}
            for load_name in ["total", "total_intercos", *episode.load_names]
        ]
        value = "total"  # episode.load_names[0]

    return options, value


def environment_ts_data(kind, episode, equipments):
    """
    Get the selected kind of timeserie trace for an equipment used in episode.

    There's four kind of trace possible:
        - Load
        - Production
        - Hazards
        - Maintenances

    :param kind: Type of trace
    :param episode: Episode studied
    :param equipments: A equipment to analyze like substation etc.
    :param prod_types: Different types of production
    :return: A list of plotly object corresponding to a trace
    """
    if kind == "Load":
        return EpisodeTrace.get_load_trace_per_equipment(episode, equipments)
    if kind == "Production":
        prod_types = episode.get_prod_types()
        return EpisodeTrace.get_all_prod_trace(episode, prod_types, equipments)
    if kind == "Hazards":
        return EpisodeTrace.get_hazard_trace(episode, equipments)
    if kind == "Maintenances":
        return EpisodeTrace.get_maintenance_trace(episode, equipments)


def agent_overflow_usage_rate_trace(episode, figure_overflow, figure_usage):
    """
    Get the trace of the overflow and the usage_rate for given episode.

    :param episode: Episode studied
    :param figure_overflow: figure which will contain the overflow trace
    :param figure_usage: figure which will contain the usage rate trace
    :returns: Plotly figure for usage_rate and for overflow
    """
    figure_overflow["data"] = episode.total_overflow_trace
    figure_usage["data"] = episode.usage_rate_trace
    return figure_overflow, figure_usage


def action_tooltip(episode_actions):
    """
    This is used to get a detailed impact action in tooltip format in order to display this tooltip on a
    plotly graph.

    :param episode_actions: episode's actions for the inspected scenario
    :return: string with action's details
    """
    tooltip = []
    # avoid reevaluation of append() see: https://wiki.python.org/moin/PythonSpeed/PerformanceTips
    tooltip_append = tooltip.append
    actions_impact = [action.impact_on_objects() for action in episode_actions]

    for action in actions_impact:
        impact_detail = []
        impact_append = impact_detail.append

        if action["has_impact"]:
            injection = action["injection"]
            force_line = action["force_line"]
            switch_line = action["switch_line"]
            topology = action["topology"]
            redispatch = action["redispatch"]

            if injection["changed"]:
                [
                    impact_append(
                        " Injection set {} to {} <br>".format(
                            detail["set"], detail["to"]
                        )
                    )
                    for detail in injection["impacted"]
                ]

            if force_line["changed"]:
                reconnections = force_line["reconnections"]
                disconnections = force_line["disconnections"]

                if reconnections["count"] > 0:
                    impact_append(
                        " Force reconnection of {} powerlines ({}) <br>".format(
                            reconnections["count"], reconnections["powerlines"]
                        )
                    )

                if disconnections["count"] > 0:
                    impact_append(
                        " Force disconnection of {} powerlines ({}) <br>".format(
                            disconnections["count"], disconnections["powerlines"]
                        )
                    )

            if switch_line["changed"]:
                impact_append(
                    " Switch status of {} powerlines ({}) <br>".format(
                        switch_line["count"], switch_line["powerlines"]
                    )
                )

            if topology["changed"]:
                bus_switch = topology["bus_switch"]
                assigned_bus = topology["assigned_bus"]
                disconnected_bus = topology["disconnect_bus"]

                if len(bus_switch) > 0:
                    [
                        impact_append(
                            " Switch bus of {} {} on substation {} <br>".format(
                                switch["object_type"],
                                switch["object_id"],
                                switch["substation"],
                            )
                        )
                        for switch in bus_switch
                    ]

                if len(assigned_bus) > 0:
                    [
                        impact_append(
                            " Assign bus {} to {} {} on substation {} <br>".format(
                                assignment["bus"],
                                assignment["object_type"],
                                assignment["object_id"],
                                assignment["substation"],
                            )
                        )
                        for assignment in assigned_bus
                    ]

                if len(disconnected_bus) > 0:
                    [
                        impact_append(
                            " Disconnect bus {} {} on substation {} <br>".format(
                                disconnection["object_type"],
                                disconnection["object_id"],
                                disconnection["substation"],
                            )
                        )
                        for disconnection in disconnected_bus
                    ]

            if redispatch["changed"]:
                generators = redispatch["generators"]
                for gen_dict in generators:
                    gen_name = gen_dict["gen_name"]
                    r_amount = gen_dict["amount"]
                    impact_append("Redispatch {} of {} <br>".format(gen_name, r_amount))

            tooltip_append("".join(impact_detail))
        else:
            tooltip_append("Do nothing")

    return tooltip


def make_action_ts(study_agent, ref_agent, scenario, layout_def=None):
    """
    Make the action timeseries trace of study and reference agents.

    :param study_agent: studied agent to compare
    :param ref_agent: reference agent to compare with
    :param scenario:
    :param layout_def: layout page
    :return: nb action and distance for each agents
    """
    ref_episode = make_episode(ref_agent, scenario)
    study_episode = make_episode(study_agent, scenario)
    actions_ts = get_actions_sum(study_episode)
    ref_agent_actions_ts = get_actions_sum(ref_episode)

    # used below to make sure the x-axis length is the study agent one
    study_agent_length = len(study_episode.action_data_table)

    action_events_df = pd.DataFrame(
        index=actions_ts.index, data=np.nan, columns=["action_events"]
    )
    action_events_df.loc[
        (actions_ts["Nb Actions"] > 0).values, "action_events"
    ] = study_episode.action_data_table.loc[
        (actions_ts["Nb Actions"] > 0).values, "distance"
    ].values
    study_text = ["<br>-".join(str(act).split("-")) for act in study_episode.actions]
    action_trace = go.Scatter(
        x=action_events_df.index,
        y=action_events_df["action_events"],
        name=study_agent + " Actions",
        mode="markers",
        marker_color="#FFEB3B",
        marker={"symbol": "hexagon", "size": 10},
        text=study_text,
    )

    ref_action_events_df = pd.DataFrame(
        index=ref_agent_actions_ts.index, data=np.nan, columns=["action_events"]
    )
    ref_action_events_df.loc[
        (ref_agent_actions_ts["Nb Actions"] > 0).values, "action_events"
    ] = ref_episode.action_data_table.loc[
        (ref_agent_actions_ts["Nb Actions"] > 0).values, "distance"
    ].values

    ref_text = ["<br>-".join(str(act).split("-")) for act in ref_episode.actions]
    ref_text = ref_text[:study_agent_length]
    ref_action_trace = go.Scatter(
        x=ref_action_events_df.index[:study_agent_length],
        y=ref_action_events_df["action_events"][:study_agent_length],
        name=ref_agent + " Actions",
        mode="markers",
        marker_color="#FF5000",
        marker={"symbol": "hexagon", "size": 10},
        text=ref_text,
    )

    distance_trace = go.Scatter(
        x=study_episode.action_data_table.timestamp,
        y=study_episode.action_data_table["distance"],
        name=study_agent,
    )

    ref_distance_trace = go.Scatter(
        x=ref_episode.action_data_table.timestamp[:study_agent_length],
        y=ref_episode.action_data_table["distance"][:study_agent_length],
        name=ref_agent,
    )

    layout_def.update(xaxis=dict(range=[distance_trace.x[0], distance_trace.x[-1]]))

    figure = {
        "data": [
            distance_trace,
            action_trace,
            ref_distance_trace,
            ref_action_trace,
        ],
        "layout": layout_def,
    }

    return figure


def make_rewards_ts(
    study_agent, ref_agent, scenario, reward_figure, cumulative_reward_figure
):
    """
    Make kpi with rewards and cumulated reward for both reference agent and study agent.

    :param study_agent: agent studied
    :param ref_agent: agent to compare with
    :param scenario:
    :param layout: display configuration
    :return: rewards and cumulated rewards for each agents
    """
    study_episode = make_episode(study_agent, scenario)
    ref_episode = make_episode(ref_agent, scenario)
    actions_ts = (
        study_episode.action_data_table.set_index("timestamp")[
            ["action_line", "action_subs", "action_redisp"]
        ]
        .sum(axis=1)
        .to_frame(name="Nb Actions")
    )
    df = observation_model.get_df_computed_reward(study_episode)
    action_events_df = pd.DataFrame(
        index=df["timestep"], data=np.nan, columns=["action_events"]
    )
    action_events_df.loc[
        (actions_ts["Nb Actions"] > 0).values, "action_events"
    ] = df.loc[(actions_ts["Nb Actions"] > 0).values, "rewards"].values
    text = ["<br>-".join(str(act).split("-")) for act in study_episode.actions]
    action_trace = go.Scatter(
        x=action_events_df.index,
        y=action_events_df["action_events"],
        name="Actions",
        mode="markers",
        marker_color="#FFEB3B",
        marker={"symbol": "hexagon", "size": 10},
        text=text,
    )

    ref_reward_trace, ref_reward_cum_trace = ref_episode.reward_trace
    (
        studied_agent_reward_trace,
        studied_agent_reward_cum_trace,
    ) = study_episode.reward_trace
    reward_figure["data"] = [ref_reward_trace, studied_agent_reward_trace, action_trace]
    cumulative_reward_figure["data"] = [
        ref_reward_cum_trace,
        studied_agent_reward_cum_trace,
    ]

    for figure in [reward_figure, cumulative_reward_figure]:
        # Base on the hypothesis that the study agent trace is in position one
        # TODO: clean this up. We should not rely on the position of the study
        # agent in the traces.
        figure["layout"].update(
            {
                "xaxis": {
                    "range": [
                        reward_figure["data"][1].x[0],
                        reward_figure["data"][1].x[-1],
                    ],
                }
            }
        )

    return reward_figure, cumulative_reward_figure


def compute_windows_range(episode, center_idx, n_clicks_left, n_clicks_right):
    """
    Compute the timestamp range for the time window
    :param episode: studied episode
    :param center_idx: timestamp at the center of the range
    :param n_clicks_left: number of times user as click on enlarge left
    :param n_clicks_right: number of times user as click on enlarge right
    :return: the timesteps minimum and maximum
    """
    timestamp_range = episode.timestamps[
        max([0, (center_idx - 10 - 5 * n_clicks_left)]) : (
            center_idx + 10 + 5 * n_clicks_right
        )
    ]
    xmin = timestamp_range[0].strftime("%Y-%m-%dT%H:%M:%S")
    xmax = timestamp_range[-1].strftime("%Y-%m-%dT%H:%M:%S")

    return xmin, xmax



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['grid2viz',
 'grid2viz.src',
 'grid2viz.src.episodes',
 'grid2viz.src.kpi',
 'grid2viz.src.macro',
 'grid2viz.src.micro',
 'grid2viz.src.overview',
 'grid2viz.src.simulation',
 'grid2viz.src.utils',
 'tests']

package_data = \
{'': ['*'],
 'grid2viz': ['assets/*',
              'assets/screenshots/*',
              'data/agents/do-nothing-baseline/*',
              'data/agents/do-nothing-baseline/000/*',
              'data/agents/do-nothing-baseline/001/*',
              'data/agents/greedy-baseline/*',
              'data/agents/greedy-baseline/000/*',
              'data/agents/greedy-baseline/001/*',
              'data/agents_dev/do-nothing-baseline/*',
              'data/agents_dev/do-nothing-baseline/000/*',
              'data/agents_dev/do-nothing-baseline/001/*',
              'data/agents_dev/do-nothing-baseline/002/*',
              'data/agents_dev/greedy-baseline/*',
              'data/agents_dev/greedy-baseline/000/*',
              'data/agents_dev/greedy-baseline/001/*',
              'data/agents_dev/multiTopology-baseline/*',
              'data/agents_dev/multiTopology-baseline/000/*',
              'data/agents_dev/redispatching-baseline/*',
              'data/agents_dev/redispatching-baseline/000/*',
              'data/agents_multi_topo/multiTopology-baseline/*',
              'data/agents_multi_topo/multiTopology-baseline/000/*',
              'data/agents_test_env_1.2.3/do-nothing-baseline/*',
              'data/agents_test_env_1.2.3/do-nothing-baseline/test_chronic/*'],
 'tests': ['data/agents/do-nothing-baseline/*',
           'data/agents/do-nothing-baseline/000/*',
           'data/agents/greedy-baseline/*',
           'data/agents/greedy-baseline/000/*',
           'data/agents/multiTopology-baseline/*',
           'data/agents/multiTopology-baseline/000/*',
           'data/agents/redispatching-baseline/*',
           'data/agents/redispatching-baseline/000/*']}

install_requires = \
['Grid2Op>=1.3.1,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'colorama>=0.4.4,<0.5.0',
 'dash-antd-components>=0.0.1-rc.2,<0.0.2',
 'dash-bootstrap-components>=0.10.7,<0.11.0',
 'dash>=1.17.0,<2.0.0',
 'dill>=0.3.3,<0.4.0',
 'imageio>=2.9.0,<3.0.0',
 'jupyter-dash>=0.3.1,<0.4.0',
 'matplotlib>=3.3.3,<4.0.0',
 'numpy==1.19.3',
 'pandapower>=2.4.0,<3.0.0',
 'pathos>=0.2.7,<0.3.0',
 'seaborn>=0.11.0,<0.12.0']

entry_points = \
{'console_scripts': ['grid2viz = grid2viz.main:main']}

setup_kwargs = {
    'name': 'grid2viz',
    'version': '1.0.0a3',
    'description': 'Grid2Op Visualization companion app.',
    'long_description': '\n# Grid2Viz: The Grid2Op Visualization companion app\n\n<!--- #[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/mjothy/grid2viz/jupyter_dash?urlpath=lab)#if launching jupyter lab directly-->\nYou can launch a demo in your web navigator by running the Grid2viz_demo notebook through Binder by clicking this button. The dataset used in this demo can be found in the [Grid2viz dataset](https://github.com/marota/Grid2viz-dataset-NeurIPS-Robustness).\n\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/mjothy/grid2viz/master/?urlpath=git-pull?repo=https://github.com/marota/Grid2viz-dataset-NeurIPS-Robustness%26amp%3Burlpath=tree/../%26amp%3Burlpath=tree/Grid2Viz_demo.ipynb%3Fautodecode)<!--- 1rst urlpath to download the dataset from a new github - 2nd urlpath to get back to a parent root directory - 3rd urlpath to directly load the notebook -->\n**WARNING:** we recommend you to use Firefox or Chrome as a web navigator rather than Internet Explorer or Edge.\n---\n\nGrid2Viz is a web application that offers several interactive views into the results of Reinforcement Learning agents that ran on the [Grid2Op](https://github.com/rte-france/Grid2Op) platform.\n\n*   [1 Documentation](#documentation)\n*   [2 Installation](#installation)\n*   [3 Run the application](#run-grid2viz)\n*   [4 Getting Started](#getting-started)\n*   [5 Caching](#caching)\n*   [6 Interface](#interface)\n*   [7 Interface](#troubleshooting)\n\n## Documentation\ngo to: https://grid2viz.readthedocs.io/en/latest/\n\n## Installation\n### Requirements:\n*   Python >= 3.6\n\n#### (Optional, recommended) Step 1: Create a virtual environment\n```commandline\npip3 install -U virtualenv\npython3 -m virtualenv venv_grid2viz\n```\n\n#### Step 2: Install from source\n```commandline\nsource venv_grid2viz/bin/activate\ngit clone https://github.com/mjothy/Grid2Viz.git\ncd Grid2Viz/\npip install -U .\n```\n\n## Run Grid2Viz\n```\nusage: grid2viz [-h] [--agents_path AGENTS_PATH] [--env_path ENV_PATH]\n                [--port PORT] [--debug]\n\nGrid2Viz\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --agents_path AGENTS_PATH\n                        The path where the log of the Agents experiences are\n                        stored. (default to None to study the example agents\n                        provided with the package)\n  --env_path ENV_PATH   The path where the environment config is stored.\n                        (default to None to use the provided default\n                        environment)\n  --port PORT           The port to serve grid2viz on. (default to 8050)\n  --debug               Enable debug mode for developers. (default to False)\n```\n\nFor example:\n\n```commandline\nsource venv_grid2viz/bin/activate\ngrid2viz --port 8000\n```\n\n> **_WARNING_** Due to the caching operation the first run can take a while. All the agents present in the configuration files\nwill be computed and then registered in cache. Depending on your agents it could take between 5 to 15min. You can follow the progress in the console.\n\n## Getting started\n\nIn order to use this tool, you need to have serialized the RL process of grid2op. The expected file system is :\n- root_dir\n    - agent_1\n        - scenario_1\n        - scenario_2\n    - agent_2\n        - scenario_1\n        - scenario_2\n        - scenario_3\n\nEach of the scenario_* files have to contain all files given by serialisation of your RL through grid2op.\nIn order to add a new agent to the app, you will have to add the agent\'s folder to this root_dir\nFor the update process of this folder chain, see the section `Caching` (in particular, when you want to overwrite the current\nagents in root_dir with new versions with the same names)\n\nIn the config.ini of this repo:\n - `agents_dir` is the path to your agents logs data directory.\n - `env_dir` is the path to the environment configuration directory. It contains a single file :\n    - coords.csv : The csv file that lists the coordinates of nodes in the network\n\nChanging this config.ini file will require a restart of the server to update.\n\nGrid2Viz provide 2 agents with a scenario for one day and for one month available in `/grid2viz/data/agents` folder:\n\n- do-nothing-baseline\n- greedy-baseline\n\nBy default the config.ini is targeting these agents as well as the environment configuration folders.\n\n##  Caching\n\nThe cache system allows you to only compute long calculations of the app once per agent/scenario.\nThe app will create a folder `_cache` in the `base_dir` of the config.ini which will contain these long calculations serialized.\n\nIf you add a new folder in your `base_dir` (either an agent, or a scenario) you will have to restart the server so the app\nreads the folder tree again.\n\n**_WARNING_** : If you overwrite the agents while they were already cached, you will have to manually reset the cache so the app\nknows to compute everything again with the updated data. To do so, you just need to delete the `_cache` folder.\n\n## Interface\n#### Scenario Selection\nThis page display up to 15 scenarios with for each one a brief summary using the best agent\'s performances.\n\n![scenario selection](https://raw.githubusercontent.com/mjothy/grid2viz/master/grid2viz/assets/screenshots/scenario_selection.png "Scenario Selection")\n\n\n#### Scenario Overview\nOn this page are displayed the best agent\'s kpi to see his performances. It\'s also here that you can select an agent that will\nbe used as reference agent in the other pages to compare to the studied agents.\n\n![scenario overview](https://raw.githubusercontent.com/mjothy/grid2viz/master/grid2viz/assets/screenshots/scenario_overview.png "Scenario Overview")\n\n#### Agent Overview\nHere\'s displayed your reference agent\'s performances. You can select an agent to study to compare it with your reference via the\ndropdown on the page. The study agent selected will be used as study agent on the last page.\n\nIn the *"instant and cumulated reward"* graph you can point timestep that will be use in the next page to study\naction in a specific timestep area.\n\n![agent overview](https://raw.githubusercontent.com/mjothy/grid2viz/master/grid2viz/assets/screenshots/agent_overview.png "Agent Overview")\n\n\n#### Agent Study\nThe Agent Study page will display kpi of your reference agent compared to your study agent on your selected timestep area.\nYou will also see a summary of the previous page\'s kpi.\n\n![agent study](https://raw.githubusercontent.com/mjothy/grid2viz/master/grid2viz/assets/screenshots/agent_study.png "Agent Study")\n\n## Run the tests\n\nTo run the tests, execute the following command:\n\n```commandline\npython3 -m unittest discover --start-directory tests --buffer\n```\n\n## Limitations\nThe app is still missing a couple features, namely a graph for visualising the flow through time, and the last line of the last screen, which will show all informations regarding the actions and observations at the selected timestep.\n\nThe Actions KPIs and the distances as well as the topological action cluster "object changed" is in alpha feature. We will need some new features from the core API to finish these features.\n\n## Troubleshooting\n### MacOS\nSome mac users have been experimenting issues when lauching the app, raising the following message:\n\n`socket.gaierror: [Errno 8] nodename nor servname provided, or not known`\n\nThe following steps might help you to overcome the issue:\n\n1. Open your terminal\n2. Type `echo $HOST` and copy the results\n3. Open the file `/etc/hosts` and make sure you include: <br>\n `127.0.0.1 PASTE RESULTS FROM echo $HOST`\n4. Save it and close it\n5. Launch grid2viz\n\n\n',
    'author': 'Mario Jothy',
    'author_email': 'mario.jothy@artelys.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import datetime
from collections import namedtuple
from pathlib import Path

import dash_antd_components as dac
import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
import dash_table as dt
import plotly.graph_objects as go

from grid2viz.src.manager import grid2viz_home_directory
from grid2viz.src.manager import make_episode, make_network_agent_study, best_agents
from grid2viz.src.utils.constants import DONT_SHOW_FILENAME
from grid2viz.src.utils.layout_helpers import modal, should_help_open

layout_def = {
    "legend": {"orientation": "h"},
    "margin": {"l": 0, "r": 0, "t": 0, "b": 0},
}

SliderParams = namedtuple("SliderParams", ["min", "max", "marks", "value"])


def indicator_line():
    return html.Div(
        id="indicator_line_id",
        className="lineBlock card",
        children=[
            html.H4("Indicators"),
            html.Div(
                className="card-body row",
                children=[
                    html.Div(
                        className="col-6",
                        children=[
                            html.H6(
                                className="text-center",
                                children="Rewards instant + cumulated for 2 agent",
                            ),
                            dbc.Tabs(
                                children=[
                                    dbc.Tab(
                                        label="Instant Reward",
                                        children=[
                                            dcc.Graph(
                                                id="rewards_ts",
                                                figure=go.Figure(
                                                    layout=layout_def,
                                                ),
                                            )
                                        ],
                                    ),
                                    dbc.Tab(
                                        label="Cumulated Reward",
                                        children=[
                                            dcc.Graph(
                                                id="cumulated_rewards_ts",
                                                figure=go.Figure(
                                                    layout=layout_def,
                                                ),
                                            )
                                        ],
                                    ),
                                ]
                            ),
                        ],
                    ),
                    html.Div(
                        className="col-6",
                        children=[
                            html.H6(
                                className="text-center",
                                children="Distance from reference grid configuration",
                            ),
                            dcc.Graph(
                                id="actions_ts",
                                figure=go.Figure(
                                    layout=layout_def, data=[dict(type="scatter")]
                                ),
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )


def flux_inspector_line(network_graph=None, slider_params=None):
    return html.Div(
        id="flux_inspector_line_id",
        className="lineBlock card",
        children=[
            html.H4("State evolution given agent actions"),
            html.Div(
                className="card-body row",
                children=[
                    html.Div(
                        className="col",
                        children=[
                            html.Div(
                                className="row",
                                children=[
                                    html.Div(
                                        className="col",
                                        children=[
                                            html.H6(
                                                className="text-center",
                                                children="Grid State evolution overtime & highlighted subs with action (yellow) - with 2 nodes (green) ",
                                            ),
                                            dcc.Graph(
                                                id="interactive_graph",
                                                figure=network_graph,
                                            ),
                                            dcc.Slider(
                                                id="slider",
                                                min=slider_params.min,
                                                max=slider_params.max,
                                                step=None,
                                                marks=slider_params.marks,
                                                value=slider_params.value,
                                            ),
                                            html.P(
                                                id="tooltip_table_micro",
                                                className="more-info-table",
                                                children=[
                                                    "Click on a row to have more info on the action"
                                                ],
                                            ),
                                        ],
                                    )
                                ],
                            ),
                            html.Div(
                                className="row",
                                children=[
                                    html.Div(
                                        className="col",
                                        children=[
                                            html.H6(
                                                className="text-center",
                                                children="Voltage, Flow and Redispatch time series",
                                            ),
                                            dac.Radio(
                                                options=[
                                                    {"label": "Flow", "value": "flow"},
                                                    {
                                                        "label": "Voltage (V)",
                                                        "value": "voltage",
                                                    },
                                                    {
                                                        "label": "Redispatch (MW)",
                                                        "value": "redispatch",
                                                    },
                                                ],
                                                value="flow",
                                                id="voltage_flow_choice",
                                                buttonStyle="solid",
                                            ),
                                            dac.Radio(
                                                options=[
                                                    {
                                                        "label": "Active Flow (MW)",
                                                        "value": "active_flow",
                                                    },
                                                    {
                                                        "label": "Current Flow (A)",
                                                        "value": "current_flow",
                                                    },
                                                    {
                                                        "label": "Flow Usage Rate",
                                                        "value": "flow_usage_rate",
                                                    },
                                                ],
                                                value="active_flow",
                                                id="flow_radio",
                                                style={"display": "none"},
                                            ),
                                            dac.Select(
                                                id="line_side_choices",
                                                options=[],
                                                value=[],
                                                mode="multiple",
                                                showArrow=True,
                                            ),
                                            dcc.Graph(
                                                id="voltage_flow_graph",
                                                figure=go.Figure(
                                                    layout=layout_def,
                                                    data=[dict(type="scatter")],
                                                ),
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        ],
                    )
                ],
            ),
        ],
    )


def context_inspector_line(best_episode, study_episode):
    return html.Div(
        id="context_inspector_line_id",
        className="lineBlock card ",
        children=[
            html.H4("Context"),
            html.Div(
                className="card-body col row",
                children=[
                    html.Div(
                        className="col-xl-5",
                        children=[
                            html.H5("Environment Time Series"),
                            dac.Radio(
                                options=[
                                    {"label": "Production", "value": "Production"},
                                    {"label": "Load", "value": "Load"},
                                    {"label": "Hazards", "value": "Hazards"},
                                    {"label": "Maintenances", "value": "Maintenances"},
                                ],
                                value="Production",
                                id="environment_choices_buttons",
                                buttonStyle="solid",
                            ),
                            dac.Select(
                                id="asset_selector",
                                options=[
                                    {"label": prod_name, "value": prod_name}
                                    for prod_name in best_episode.prod_names
                                ],
                                value="solar",
                                # episode.prod_names[3],#[episode.prod_names[0],episode.prod_names[1]],#[prod_name for prod_name in episode.prod_names if prod_name in ['wind','solar']],#episode.prod_names[0]
                                mode="multiple",
                                showArrow=True,
                            ),
                            dcc.Graph(
                                id="env_charts_ts",
                                style={"margin-top": "1em"},
                                figure=go.Figure(
                                    layout=layout_def, data=[dict(type="scatter")]
                                ),
                                config=dict(displayModeBar=False),
                            ),
                        ],
                    ),
                    html.Div(
                        className="col-xl-7",
                        children=[
                            html.H5("Studied agent Metrics"),
                            html.Div(
                                className="row",
                                children=[
                                    html.Div(
                                        className="col-6",
                                        children=[
                                            html.H5(
                                                "Usage rate", className="text-center"
                                            ),
                                            dcc.Graph(
                                                id="usage_rate_ts",
                                                style={"margin-top": "1em"},
                                                figure=go.Figure(
                                                    layout=layout_def,
                                                    data=study_episode.usage_rate_trace,
                                                ),
                                                config=dict(displayModeBar=False),
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="col-6",
                                        children=[
                                            html.H5(
                                                "Overflow", className="text-center"
                                            ),
                                            dcc.Graph(
                                                id="overflow_ts",
                                                style={"margin-top": "1em"},
                                                figure=go.Figure(
                                                    layout=layout_def,
                                                    data=study_episode.total_overflow_trace,
                                                ),
                                                config=dict(displayModeBar=False),
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )


all_info_line = html.Div(
    id="all_info_line_id",
    className="lineBlock card ",
    children=[
        html.H4("All"),
        html.Div(
            className="card-body col row",
            children=[
                html.Div(
                    className="col-xl-10",
                    children=[
                        dt.DataTable(
                            id="all_info_table",
                            filter_action="native",
                            sort_action="native",
                            sort_mode="multi",
                            page_action="native",
                            page_current=0,
                            page_size=20,
                        )
                    ],
                )
            ],
        ),
    ],
)


def center_index(user_selected_timestamp, episode):
    if user_selected_timestamp is not None:
        center_indx = episode.timestamps.index(
            datetime.datetime.strptime(user_selected_timestamp, "%Y-%m-%d %H:%M")
        )
    else:
        center_indx = 0
    return center_indx


def slider_params(user_selected_timestamp, episode):
    # min max marks value
    value = center_index(user_selected_timestamp, episode)
    n_clicks_left = 0
    n_clicks_right = 0
    min_ = max([0, (value - 10 - 5 * n_clicks_left)])
    max_ = min([(value + 10 + 5 * n_clicks_right), len(episode.timestamps)])
    timestamp_range = episode.timestamps[min_:max_]
    timestamp_range = [timestamp.time() for timestamp in timestamp_range]
    marks = dict(enumerate(timestamp_range))
    return SliderParams(min_, max_, marks, value)


def layout(user_selected_timestamp, study_agent, ref_agent, scenario):
    best_episode = make_episode(best_agents[scenario]["agent"], scenario)
    new_episode = make_episode(study_agent, scenario)
    center_indx = center_index(user_selected_timestamp, new_episode)
    network_graph = make_network_agent_study(new_episode, timestep=center_indx)

    open_help = should_help_open(
        Path(grid2viz_home_directory) / DONT_SHOW_FILENAME("micro")
    )
    header = "Analyze further your agent"
    body = (
        "Select a time step in the navbar dropdown and analyze what happened "
        "at that time to understand the agent behavior."
    )
    return html.Div(
        id="micro_page",
        children=[
            indicator_line(),
            flux_inspector_line(
                network_graph, slider_params(user_selected_timestamp, new_episode)
            ),
            context_inspector_line(best_episode, new_episode),
            all_info_line,
            modal(id_suffix="micro", is_open=open_help, header=header, body=body),
        ],
    )



from collections import namedtuple
from pathlib import Path

import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
import dash_table as dt
import plotly.graph_objects as go

from grid2viz.src.kpi import actions_model
from grid2viz.src.manager import (
    make_episode,
    agents,
    make_network_agent_overview,
    best_agents,
    grid2viz_home_directory,
)
from grid2viz.src.utils.constants import DONT_SHOW_FILENAME
from grid2viz.src.utils.graph_utils import layout_def, layout_no_data, max_or_zero
from grid2viz.src.utils.layout_helpers import modal, should_help_open


def indicator_line(scenario, study_agent, ref_agent):
    episode = make_episode(study_agent, scenario)
    ref_episode = make_episode(ref_agent, scenario)
    figures_distribution = action_distrubtion(episode, ref_episode)

    network_graph = make_network_agent_overview(episode)

    nb_actions = episode.action_data_table[
        ["action_line", "action_subs", "action_redisp"]
    ].sum()

    if nb_actions.sum() == 0:
        pie_figure = go.Figure(layout=layout_no_data("No Actions for this Agent"))
    else:
        pie_figure = go.Figure(
            layout=layout_def,
            data=[
                go.Pie(
                    labels=[
                        "Actions on Lines",
                        "Actions on Substations",
                        "Redispatching Actions",
                    ],
                    values=[
                        nb_actions["action_line"],
                        nb_actions["action_subs"],
                        nb_actions["action_redisp"],
                    ],
                )
            ],
        )

    return html.Div(
        className="lineBlock card",
        children=[
            html.H4("Indicators"),
            html.Div(
                className="card-body row",
                children=[
                    html.Div(
                        className="col-2",
                        children=[
                            html.H5("Study Agent Summary"),
                            html.Div(
                                className="m-2",
                                children=[
                                    html.P(
                                        id="indicator_score_output",
                                        className="border-bottom h3 mb-0 text-right",
                                        children=round(
                                            episode.meta["cumulative_reward"]
                                        ),
                                    ),
                                    html.P(className="text-muted", children="Score"),
                                ],
                            ),
                            html.Div(
                                className="m-2",
                                children=[
                                    html.P(
                                        id="indicator_nb_maintenances",
                                        className="border-bottom h3 mb-0 text-right",
                                        children="{}/{}".format(
                                            episode.meta["nb_timestep_played"],
                                            episode.meta["chronics_max_timestep"],
                                        ),
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="Agent's Survival",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="m-2",
                                children=[
                                    html.P(
                                        id="indicator_nb_overflow",
                                        className="border-bottom h3 mb-0 text-right",
                                        children=episode.total_overflow_ts[
                                            "value"
                                        ].sum(),
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="Number of Overflow",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="m-2",
                                children=[
                                    html.P(
                                        id="indicator_nb_action",
                                        className="border-bottom h3 mb-0 text-right",
                                        children=episode.action_data_table[
                                            ["action_line", "action_subs"]
                                        ]
                                        .sum(axis=1)
                                        .sum(),
                                    ),
                                    html.P(
                                        className="text-muted ",
                                        children="Number of Action",
                                    ),
                                ],
                            ),
                            html.Div(
                                className="m-2",
                                children=[
                                    html.P(
                                        id="indicator_nb_maintenances",
                                        className="border-bottom h3 mb-0 text-right",
                                        children=episode.nb_maintenances,
                                    ),
                                    html.P(
                                        className="text-muted",
                                        children="Number of Maintenances",
                                    ),
                                ],
                            ),
                        ],
                    ),
                    html.Div(
                        className="col-3",
                        children=[
                            html.H6(
                                className="text-center",
                                children="Type Action Repartition",
                            ),
                            dcc.Graph(id="agent_study_pie_chart", figure=pie_figure),
                        ],
                    ),
                    html.Div(
                        className="col-7",
                        children=[
                            html.H6(
                                className="text-center",
                                children="Impacted grid assets: attacks (dash orange) & overflow (red) and subs with action",
                            ),
                            dcc.Graph(id="network_actions", figure=network_graph),
                        ],
                    ),
                    html.Div(
                        className="col-12",
                        children=[
                            html.H2(
                                className="text-center",
                                children="Actions Distributions",
                            )
                        ],
                    ),
                    html.Div(
                        className="col-4",
                        children=[
                            html.H6(className="text-center", children="On Substations"),
                            dcc.Graph(
                                id="distribution_substation_action_chart",
                                figure=figures_distribution.on_subs,
                            ),
                        ],
                    ),
                    html.Div(
                        className="col-4",
                        children=[
                            html.H6(className="text-center", children="On Lines"),
                            dcc.Graph(
                                id="distribution_line_action_chart",
                                figure=figures_distribution.on_lines,
                            ),
                        ],
                    ),
                    html.Div(
                        className="col-4",
                        children=[
                            html.H6(className="text-center", children="Redispatching"),
                            dcc.Graph(
                                id="distribution_redisp_action_chart",
                                figure=figures_distribution.redisp,
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )


def overview_line(timestamps=None, from_scenario_selection=True):
    if timestamps is None or from_scenario_selection:
        timestamps = []
    return html.Div(
        id="overview_line_id",
        className="lineBlock card",
        children=[
            html.H4("Overview"),
            html.Div(
                className="card-body row",
                children=[
                    html.Div(
                        className="col-2",
                        children=[
                            html.H4("Agent study Timestep selection"),
                            dt.DataTable(
                                id="timeseries_table",
                                columns=[{"name": "Timestamps", "id": "Timestamps"}],
                                style_as_list_view=True,
                                row_deletable=True,
                                data=timestamps,
                                filter_action="native",
                                sort_action="native",
                                style_table={
                                    "overflow-y": "scroll",
                                    # 'width': 'auto',
                                    "height": "100%",
                                },
                            ),
                            html.Div(
                                html.P(
                                    'Interactively select a timestep by clicking on the "Instant and Accumulated Reward" '
                                    "Time Serie to study it on the next page"
                                ),
                                className="mt-1",
                            ),
                        ],
                    ),
                    html.Div(
                        className="col-10",
                        children=[
                            html.Div(
                                className="row",
                                children=[
                                    html.Div(
                                        className="col-6",
                                        children=[
                                            html.H6(
                                                className="text-center",
                                                children="Instant and Cumulated Reward",
                                            ),
                                            dbc.Tabs(
                                                children=[
                                                    dbc.Tab(
                                                        label="Instant Reward",
                                                        children=[
                                                            dcc.Graph(
                                                                id="rewards_timeserie",
                                                                figure=go.Figure(
                                                                    layout=layout_def,
                                                                ),
                                                            )
                                                        ],
                                                    ),
                                                    dbc.Tab(
                                                        label="Cumulated Reward",
                                                        children=[
                                                            dcc.Graph(
                                                                id="cumulated_rewards_timeserie",
                                                                figure=go.Figure(
                                                                    layout=layout_def,
                                                                ),
                                                            )
                                                        ],
                                                    ),
                                                ]
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="col-6",
                                        children=[
                                            html.H6(
                                                className="text-center",
                                                children="Overflow, Maintenances, Hazards and Attacks",
                                            ),
                                            dcc.Graph(
                                                id="overflow_graph_study",
                                                figure=go.Figure(
                                                    layout=layout_def,
                                                    data=[dict(type="scatter")],
                                                ),
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                            html.Div(
                                className="row",
                                children=[
                                    html.Div(
                                        className="col-6",
                                        children=[
                                            html.H6(
                                                className="text-center",
                                                children="Distance from reference grid configuration",
                                            ),
                                            dcc.Graph(
                                                id="action_timeserie",
                                                figure=go.Figure(
                                                    layout=layout_def,
                                                    data=[dict(type="scatter")],
                                                ),
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="col-6",
                                        children=[
                                            html.H6(
                                                className="text-center",
                                                children="Usage Rate",
                                            ),
                                            dcc.Graph(
                                                id="usage_rate_graph_study",
                                                figure=go.Figure(
                                                    layout=layout_def,
                                                    data=[dict(type="scatter")],
                                                ),
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )


def inspector_line(study_agent, scenario):
    new_episode = make_episode(study_agent, scenario)
    cols, data = get_table(new_episode)

    data_table = dt.DataTable(
        columns=cols,
        data=data,
        id="inspector_datable",
        filter_action="native",
        sort_action="native",
        sort_mode="multi",
        page_action="native",
        page_current=0,
        page_size=7,
    )

    return html.Div(
        className="lineBlock card ",
        children=[
            html.H4("Inspector For Study Agent"),
            html.Div(
                className="container-fluid",
                id="action_table_div",
                children=[
                    html.Div(
                        className="row",
                        children=[
                            html.Div(
                                className="col flex-center", children=[data_table]
                            ),
                        ],
                    ),
                    html.Div(
                        className="row",
                        children=[
                            html.Div(
                                className="col flex-center",
                                children=[
                                    html.Label(
                                        children=[
                                            "The documentation for the filtering syntax can be found ",
                                            html.A(
                                                "here.",
                                                href="https://dash.plot.ly/datatable/filtering",
                                                target="_blank",
                                            ),
                                        ]
                                    ),
                                ],
                            )
                        ],
                    ),
                    html.Div(
                        className="row",
                        children=[
                            html.Div(
                                className="col",
                                children=[
                                    html.P(
                                        id="tooltip_table",
                                        className="more-info-table",
                                        children=[
                                            "Click on a row to have more info on the action"
                                        ],
                                    )
                                ],
                            )
                        ],
                    ),
                ],
            ),
        ],
    )


def get_table(episode):
    table = actions_model.get_action_table_data(episode)
    table["id"] = table["timestep"]
    table["timestep_reward"] = table["timestep_reward"].map(
        lambda x: "{:,.2f}".format(float("".join(str(x).split(","))))
    )
    table.set_index("id", inplace=True, drop=False)
    cols_to_exclude = ["id", "lines_modified", "subs_modified"]
    return [
        {"name": i, "id": i} for i in table.columns if i not in cols_to_exclude
    ], table.to_dict("record")


ActionsDistribution = namedtuple(
    "ActionsDistribution", ["on_subs", "on_lines", "redisp"]
)


def action_distrubtion(episode, ref_episode):
    actions_per_sub = actions_model.get_action_per_sub(episode)
    actions_per_sub.append(actions_model.get_action_per_sub(ref_episode)[0])
    y_max = None

    if len(actions_per_sub[0]["y"]) == 0:
        figure_subs = go.Figure(
            layout=layout_no_data("No Actions on subs for this Agent")
        )
    else:
        figure_subs = go.Figure(layout=layout_def, data=actions_per_sub)
        y_max = max(map(max_or_zero, [trace.y for trace in actions_per_sub])) + 1

    actions_per_lines = actions_model.get_action_per_line(episode)
    actions_per_lines.append(actions_model.get_action_per_line(ref_episode)[0])

    if len(actions_per_lines[0]["y"]) == 0:
        figure_lines = go.Figure(
            layout=layout_no_data("No Actions on lines for this Agent")
        )
    else:
        figure_lines = go.Figure(layout=layout_def, data=actions_per_lines)
        if y_max is None:
            y_max = max(map(max_or_zero, [trace.y for trace in actions_per_lines])) + 1
        if max(map(max_or_zero, [trace.y for trace in actions_per_lines])) > y_max:
            y_max = max(map(max_or_zero, [trace.y for trace in actions_per_lines])) + 1

    actions_redisp = actions_model.get_action_redispatch(episode)
    actions_redisp.append(actions_model.get_action_redispatch(ref_episode)[0])

    if len(actions_redisp[0]["y"]) == 0:
        figure_redisp = go.Figure(
            layout=layout_no_data("No redispatching actions for this Agent")
        )
    else:
        figure_redisp = go.Figure(layout=layout_def, data=actions_redisp)
        if y_max is None:
            y_max = max(map(max_or_zero, [trace.y for trace in actions_redisp])) + 1
        if max(map(max_or_zero, [trace.y for trace in actions_redisp])) > y_max:
            y_max = max(map(max_or_zero, [trace.y for trace in actions_redisp])) + 1

    if y_max:
        figure_subs.update_yaxes(range=[0, y_max])
        figure_lines.update_yaxes(range=[0, y_max])
        figure_redisp.update_yaxes(range=[0, y_max])

    return ActionsDistribution(
        on_subs=figure_subs, on_lines=figure_lines, redisp=figure_redisp
    )


def layout(timestamps, scenario, study_agent, ref_agent, from_scenario_selection):
    if study_agent is None:
        study_agent = best_agents[scenario]["agent"]
    open_help = should_help_open(
        Path(grid2viz_home_directory) / DONT_SHOW_FILENAME("macro")
    )
    header = "Take a look at your agent"
    body = (
        "Select an agent to study in the dropdown menu and analyse it with "
        "respect to the reference agent. Click on the reward graph to select "
        "some time steps to study further your agent. When you have selected time steps, "
        "go on to the Study agent view."
    )

    return html.Div(
        id="overview_page",
        children=[
            dcc.Store(id="relayoutStoreMacro"),
            indicator_line(scenario, study_agent, ref_agent),
            overview_line(timestamps, from_scenario_selection),
            inspector_line(study_agent, scenario),
            modal(id_suffix="macro", is_open=open_help, header=header, body=body),
        ],
    )



from pathlib import Path

import dash_html_components as html
import dash_bootstrap_components as dbc


def modal(id_suffix: str = "", is_open: bool = True, header: str = "", body: str = ""):
    """
    Creat a modal used in every view to guide the user.
    Parameters
    ----------
    id_suffix: `str`
        Suffix used to make ids unique
    is_open: `bool`
        Control the visibility of the modal
    header: `str`
        Text for the header of the modal
    body: `str`
        Text for the body fo the modal
    Returns
    -------
    The dash component
    """
    id_modal = f"modal_{id_suffix}"
    id_close_btn = f"close_{id_suffix}"
    id_dont_show_again = f"dont_show_again_{id_suffix}"
    id_dont_show_again_div = f"dont_show_again_div_{id_suffix}"
    id_close_area = f"close_area_{id_suffix}"
    id_image = f"modal_image_{id_suffix}"

    return html.Div(
        [
            dbc.Modal(
                [
                    dbc.ModalHeader(header),
                    dbc.ModalBody(
                        children=[
                            body,
                            html.Div(
                                id=id_close_area,
                                children=[
                                    html.Div(
                                        id=id_dont_show_again_div,
                                        children=[
                                            dbc.Checkbox(
                                                id=id_dont_show_again,
                                                className="mt-4",
                                            ),
                                            dbc.Label(
                                                "Do not show again",
                                                html_for=id_dont_show_again,
                                                className="mr-1 mt-3 ml-1 pt-1",
                                            ),
                                        ],
                                    ),
                                    dbc.Button(
                                        "Close",
                                        id=id_close_btn,
                                        color="primary",
                                        className="p-2 m-2",
                                    ),
                                ],
                                className="d-flex justify-content-end bg-light",
                            ),
                            dbc.Card(dbc.CardImg(id=id_image)),
                        ]
                    ),
                    dbc.ModalFooter(children=[]),
                ],
                id=id_modal,
                is_open=is_open,
                size="xl",
            ),
        ]
    )


def should_help_open(filepath_to_check: str):
    open_modal = True
    if Path(filepath_to_check).is_file():
        open_modal = False
    return open_modal



import pandas as pd

from .env_actions import env_actions


def get_prod_and_conso(episode):
    episode_data = episode
    productions = pd.pivot_table(
        episode_data.production,
        index="timestamp",
        values="value",
        columns=["equipment_name"],
    )

    loads = pd.pivot_table(
        episode_data.load, index="timestamp", values="value", columns=["equipment_name"]
    )

    prods_and_loads = productions.merge(loads, left_index=True, right_index=True)
    return prods_and_loads


def get_episode_active_consumption_ts(episode):
    return [sum(obs.load_p) for obs in episode.observations]


def get_prod(episode, equipments=None):
    if equipments is None:
        return episode.production
    else:
        return episode.production.loc[
            episode.production.equipment_name.isin(equipments), :
        ]


def get_load(episode, equipments=None):
    if equipments is None:
        return episode.load
    else:
        return episode.load.loc[episode.load.equipment_name.isin(equipments), :]


def get_rho(episode):
    return episode.rho


def get_df_computed_reward(episode):
    return episode.computed_reward


# quantiles utilities
def quantile10(df):
    return df.quantile(0.1)


def quantile25(df):
    return df.quantile(0.25)


def quantile75(df):
    return df.quantile(0.75)


def quantile90(df):
    return df.quantile(0.90)


def get_usage_rate(episode):
    rho = get_rho(episode)
    return (
        rho.groupby("timestamp")
        .aggregate(["median", quantile10, quantile25, quantile75, quantile90, "max"])[
            ["value"]
        ]
        .reset_index()
    )


def init_table_inspection_data(episode):
    ts_hazards = env_actions(episode, which="hazards", kind="ts", aggr=True)
    ts_maintenances = env_actions(episode, which="maintenances", kind="ts", aggr=True)
    table = ts_hazards.merge(ts_maintenances, left_index=True, right_index=True)
    table = table.reset_index()
    if not table.empty:
        table["IsWorkingDay"] = table["timestamp"].dt.weekday < 5
    return table



import datetime as dt

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

"""
WARNING :
These imports are mandatory to build the dependance tree and actually add the callbacks to the dash decoration routine
Do not remove !
The "as ..." are also mandatory, other nothing is done.
"""
from grid2viz.src.overview import overview_lyt as overview
from grid2viz.src.macro import macro_lyt as macro
from grid2viz.src.micro import micro_lyt as micro
from grid2viz.src.episodes import episodes_lyt as episodes

"""
End Warning
"""
from grid2viz.src import manager
from grid2viz.src.utils import common_graph


def center_index(user_selected_timestamp, episode):
    if user_selected_timestamp is not None:
        center_indx = episode.timestamps.index(
            dt.datetime.strptime(user_selected_timestamp, "%Y-%m-%d %H:%M")
        )
    else:
        center_indx = 0
    return center_indx


def compute_window(user_selected_timestamp, study_agent, scenario):
    if user_selected_timestamp is not None:
        n_clicks_left = 0
        n_clicks_right = 0
        new_episode = manager.make_episode(study_agent, scenario)
        center_indx = center_index(user_selected_timestamp, new_episode)

        return common_graph.compute_windows_range(
            new_episode, center_indx, n_clicks_left, n_clicks_right
        )


def register_callbacks_main(app):
    @app.callback(
        [
            Output("page-content", "children"),
            Output("page", "data"),
            Output("nav_scen_select", "active"),
            Output("nav_scen_over", "active"),
            Output("nav_agent_over", "active"),
            Output("nav_agent_study", "active"),
            Output("reset_timeseries_table_macro", "data"),
        ],
        [Input("url", "pathname")],
        [
            State("scenario", "data"),
            State("agent_ref", "data"),
            State("agent_study", "data"),
            State("user_timestamps", "value"),
            State("page", "data"),
            State("user_timestamps_store", "data"),
            State("reset_timeseries_table_macro", "data"),
        ],
    )
    def register_page_lyt(
        pathname,
        scenario,
        ref_agent,
        study_agent,
        user_selected_timestamp,
        prev_page,
        timestamps_store,
        reset_ts_table_macro,
    ):
        if timestamps_store is None:
            timestamps_store = []
        timestamps = [
            dict(Timestamps=timestamp["label"]) for timestamp in timestamps_store
        ]

        # When you have a proxy, like on Binder, the pathname can be long.
        # So we split it and only keep the last piece
        pathName_split = pathname.split("/")
        pathName_split = pathName_split[len(pathName_split) - 1]

        if pathName_split and pathName_split[1:] == prev_page:
            raise PreventUpdate

        if prev_page == "episodes":
            reset_ts_table_macro = True

        if pathName_split == "episodes" or pathName_split == "" or not pathName_split:
            return (
                episodes.layout(),
                "episodes",
                True,
                False,
                False,
                False,
                reset_ts_table_macro,
            )
        elif pathName_split == "overview":
            return (
                overview.layout(scenario, ref_agent),
                "overview",
                False,
                True,
                False,
                False,
                reset_ts_table_macro,
            )
        elif pathName_split == "macro":
            if ref_agent is None:
                raise PreventUpdate
            return (
                macro.layout(
                    timestamps, scenario, study_agent, ref_agent, reset_ts_table_macro
                ),
                "macro",
                False,
                False,
                True,
                False,
                False,
            )
        elif pathName_split == "micro":
            if ref_agent is None or study_agent is None:
                raise PreventUpdate
            layout = (
                micro.layout(user_selected_timestamp, study_agent, ref_agent, scenario),
                "micro",
                False,
                False,
                False,
                True,
                reset_ts_table_macro,
            )
            return layout
        else:
            return 404, ""

    @app.callback(Output("scen_lbl", "children"), [Input("scenario", "data")])
    def update_scenario_label(scenario):
        if scenario is None:
            scenario = ""
        return scenario

    @app.callback(
        Output("ref_ag_lbl", "children"), [Input("select_ref_agent", "value")]
    )
    def update_ref_agent_label(agent):
        if agent is None:
            agent = ""
        return agent

    @app.callback(Output("study_ag_lbl", "children"), [Input("agent_study", "data")])
    def update_study_agent_label(agent):
        if agent is None:
            agent = ""
        return agent

    @app.callback(Output("user_timestamp_div", "className"), [Input("url", "pathname")])
    def show_user_timestamps(pathname):
        pathName_split = pathname.split("/")
        pathName_split = pathName_split[len(pathName_split) - 1]

        class_name = "ml-4 row"
        if pathName_split != "micro":
            class_name = " ".join([class_name, "hidden"])
        return class_name

    @app.callback(
        Output("user_timestamps", "options"),
        [Input("user_timestamps_store", "data"), Input("agent_study", "data")],
        [State("scenario", "data")],
    )
    def update_user_timestamps_options(data, agent, scenario):
        if data is None:
            raise PreventUpdate
        episode = manager.make_episode(agent, scenario)
        nb_timesteps_played = episode.meta["nb_timestep_played"]
        return [
            ts
            for ts in data
            if dt.datetime.strptime(ts["value"], "%Y-%m-%d %H:%M")
            in episode.timestamps[:nb_timesteps_played]
        ]

    @app.callback(
        Output("user_timestamps", "value"),
        [Input("user_timestamps_store", "data"), Input("agent_study", "data")],
        [State("scenario", "data")],
    )
    def update_user_timestamps_value(data, agent, scenario):
        if not data:
            raise PreventUpdate
        episode = manager.make_episode(agent, scenario)
        nb_timesteps_played = episode.meta["nb_timestep_played"]
        filtered_data = [
            ts
            for ts in data
            if dt.datetime.strptime(ts["value"], "%Y-%m-%d %H:%M")
            in episode.timestamps[:nb_timesteps_played]
        ]
        if filtered_data:
            return filtered_data[0]["value"]

    @app.callback(
        Output("enlarge_left", "n_clicks"), [Input("user_timestamps", "value")]
    )
    def reset_n_cliks_left(value):
        return 0

    @app.callback(
        Output("enlarge_right", "n_clicks"), [Input("user_timestamps", "value")]
    )
    def reset_n_cliks_right(value):
        return 0

    @app.callback(
        [
            Output("select_ref_agent", "options"),
            Output("select_ref_agent", "disabled"),
            Output("select_ref_agent", "value"),
        ],
        [Input("scenario", "data"), Input("url", "pathname")],
    )
    def update_ref_agent_select_options(scenario, pathname):
        if scenario is None:
            raise PreventUpdate
        options = [
            {"label": agent, "value": agent}
            for agent in manager.agent_scenario[scenario]
        ]
        value = manager.agent_scenario[scenario][0]
        manager.make_episode(value, scenario)
        disabled = False
        pathname_split = pathname.split("/")
        pathname_split = pathname_split[len(pathname_split) - 1]
        if pathname_split in ["episodes", "micro"]:
            disabled = True
        return options, disabled, value

    @app.callback(
        [
            Output("select_study_agent", "options"),
            Output("select_study_agent", "disabled"),
            Output("select_study_agent", "value"),
        ],
        [Input("scenario", "data"), Input("url", "pathname")],
    )
    def update_ref_agent_select_options(scenario, pathname):
        if scenario is None:
            raise PreventUpdate
        options = [
            {"label": agent, "value": agent}
            for agent in manager.agent_scenario[scenario]
        ]
        value = manager.best_agents[scenario]["agent"]
        manager.make_episode(value, scenario)
        disabled = False
        pathname_split = pathname.split("/")
        pathname_split = pathname_split[len(pathname_split) - 1]
        if pathname_split in ["micro", "episodes", "overview"]:
            disabled = True
        return options, disabled, value

    @app.callback(
        Output("agent_study", "data"),
        [Input("select_study_agent", "value")],
        [State("scenario", "data")],
    )
    def update_agent_study_store(agent, scenario):
        manager.make_episode(agent, scenario)
        return agent

    @app.callback(
        Output("agent_ref", "data"),
        [Input("select_ref_agent", "value")],
        [State("scenario", "data")],
    )
    def update_agent_study_store(agent, scenario):
        manager.make_episode(agent, scenario)
        return agent

    @app.callback(
        Output("window", "data"),
        [
            Input("enlarge_left", "n_clicks"),
            Input("enlarge_right", "n_clicks"),
            Input("user_timestamps", "value"),
        ],
        [State("agent_study", "data"), State("scenario", "data")],
    )
    def compute_window(
        n_clicks_left, n_clicks_right, user_selected_timestamp, study_agent, scenario
    ):
        if user_selected_timestamp is None:
            raise PreventUpdate
        if n_clicks_left is None:
            n_clicks_left = 0
        if n_clicks_right is None:
            n_clicks_right = 0
        new_episode = manager.make_episode(study_agent, scenario)
        center_indx = new_episode.timestamps.index(
            dt.datetime.strptime(user_selected_timestamp, "%Y-%m-%d %H:%M")
        )
        return common_graph.compute_windows_range(
            new_episode, center_indx, n_clicks_left, n_clicks_right
        )
